use std::fmt::Debug;

use galactica_content::Content;
use glyphon::TextArea;
use log::debug;

use super::scenes::{UiFlyingScene, UiLandedScene, UiOutfitterScene};
use crate::{RenderInput, RenderState};

/// Output from a ui scene step
pub struct UiSceneStepResult {
	/// If Some, switch to this scene
	pub new_scene: Option<UiScenes>,
}

pub trait UiScene<'this>
where
	Self: 'this,
{
	/// Draw this scene
	fn draw(&mut self, input: &RenderInput, state: &mut RenderState);

	/// Update this scene's state for this frame.
	/// Handles clicks, keys, etc.
	fn step(&mut self, input: &RenderInput, state: &mut RenderState) -> UiSceneStepResult;

	/// Add all textareas in this scene to `v`
	fn get_textareas(
		&'this self,
		v: &mut Vec<TextArea<'this>>,
		input: &RenderInput,
		state: &RenderState,
	);
}

pub(crate) enum UiScenes {
	Landed(UiLandedScene),
	Flying(UiFlyingScene),
	Outfitter(UiOutfitterScene),
}

impl Debug for UiScenes {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		match self {
			Self::Flying(_) => write!(f, "UiScenes::Flying"),
			Self::Landed(_) => write!(f, "UiScenes::Landed"),
			Self::Outfitter(_) => write!(f, "UiScenes::Outfitter"),
		}
	}
}

/*
impl UiScenes {
	fn is_flying(&self) -> bool {
		match self {
			Self::Flying(_) => true,
			_ => false,
		}
	}

	fn is_landed(&self) -> bool {
		match self {
			Self::Landed(_) => true,
			_ => false,
		}
	}

	fn is_outfitter(&self) -> bool {
		match self {
			Self::Outfitter(_) => true,
			_ => false,
		}
	}
}
*/

impl<'a> UiScene<'a> for UiScenes {
	fn draw(&mut self, input: &RenderInput, state: &mut RenderState) {
		match self {
			Self::Flying(s) => s.draw(input, state),
			Self::Landed(s) => s.draw(input, state),
			Self::Outfitter(s) => s.draw(input, state),
		}
	}

	fn step(&mut self, input: &RenderInput, state: &mut RenderState) -> UiSceneStepResult {
		match self {
			Self::Flying(s) => s.step(input, state),
			Self::Landed(s) => s.step(input, state),
			Self::Outfitter(s) => s.step(input, state),
		}
	}

	fn get_textareas(
		&'a self,
		v: &mut Vec<TextArea<'a>>,
		input: &RenderInput,
		state: &RenderState,
	) {
		match self {
			Self::Flying(s) => s.get_textareas(v, input, state),
			Self::Landed(s) => s.get_textareas(v, input, state),
			Self::Outfitter(s) => s.get_textareas(v, input, state),
		}
	}
}

pub struct UiManager {
	current_scene: UiScenes,
}

impl UiManager {
	pub fn new(ct: &Content, state: &mut RenderState) -> Self {
		Self {
			current_scene: UiScenes::Flying(UiFlyingScene::new(ct, state)),
		}
	}

	/// Change the current scene
	pub fn set_scene(&mut self, scene: UiScenes) {
		debug!("switching to {:?}", scene);
		self.current_scene = scene;
	}

	/// Draw all ui elements
	pub fn draw(&mut self, input: &RenderInput, state: &mut RenderState) {
		loop {
			let r = self.current_scene.step(input, state);
			if let Some(new_scene) = r.new_scene {
				debug!("{:?} changed scene", self.current_scene);
				self.set_scene(new_scene)
			} else {
				break;
			}
		}

		self.current_scene.draw(input, state);
	}

	/// Textareas to show while player is flying
	pub fn get_textareas(&self, input: &RenderInput, state: &RenderState) -> Vec<TextArea> {
		let mut v = Vec::with_capacity(5);
		self.current_scene.get_textareas(&mut v, input, state);
		return v;
	}
}
