use galactica_content::{Content, SystemObject, SystemObjectHandle};
use galactica_system::{data::ShipState, phys::PhysSimShipHandle};
use glyphon::{Attrs, TextArea, Weight};

use crate::{
	ui::{
		manager::{UiScene, UiSceneStepResult, UiScenes},
		util::{UiSprite, UiTextArea},
	},
	RenderInput, RenderState,
};

use super::UiOutfitterScene;

pub struct UiLandedScene {
	// UI elements
	description: UiTextArea,
	title: UiTextArea,
	frame: UiSprite,
	landscape: UiSprite,
	button: UiSprite,

	/// What object we're displaying currently.
	/// Whenever this changes, we need to reflow text.
	current_object: Option<SystemObjectHandle>,

	/// True if we've caught a left click event.
	/// Used for edge detection.
	leftclick_down: bool,
}

impl UiLandedScene {
	pub fn new(ct: &Content, state: &mut RenderState) -> Self {
		let frame = UiSprite::from(ct, &ct.get_ui().landed_frame);
		let button = UiSprite::from(ct, &ct.get_ui().landed_button);
		let landscape = UiSprite::from_with_sprite(
			ct,
			&ct.get_ui().landed_landscape,
			ct.get_sprite_handle("ui::landscape::test"),
		);

		let s = Self {
			// height of element in logical pixels
			current_object: None,
			leftclick_down: false,

			description: UiTextArea::from(ct, state, &ct.get_ui().landed_planet_desc),
			title: UiTextArea::from(ct, state, &ct.get_ui().landed_planet_name),

			frame,
			landscape,
			button,
		};

		return s;
	}

	fn reflow(&mut self, planet: &SystemObject, state: &mut RenderState) {
		self.description.set_text(
			state,
			&planet.desc,
			Attrs::new()
				.weight(Weight::NORMAL)
				.family(glyphon::Family::SansSerif),
		);

		self.title.set_text(
			state,
			&planet.name,
			Attrs::new()
				.weight(Weight::BOLD)
				.family(glyphon::Family::Serif),
		);
		self.current_object = Some(planet.handle);
	}
}

impl<'this> UiScene<'this> for UiLandedScene {
	fn step(&mut self, input: &RenderInput, state: &mut RenderState) -> UiSceneStepResult {
		self.button.step(input, state);
		self.landscape.step(input, state);
		self.frame.step(input, state);

		let mut new_scene = None;
		if input.player.input.pressed_leftclick() && !self.leftclick_down {
			self.leftclick_down = true;
			if self.button.contains_mouse(input, state) {
				new_scene = Some(UiScenes::Outfitter(UiOutfitterScene::new(input.ct, state)));
			}
		} else if !input.player.input.pressed_leftclick() {
			self.leftclick_down = false;
		}

		return UiSceneStepResult { new_scene };
	}

	fn draw(&mut self, input: &RenderInput, state: &mut RenderState) {
		// Get required data
		let ship_handle = input.player.ship.unwrap();
		let ship_data = &input
			.phys_img
			.get_ship(&PhysSimShipHandle(ship_handle))
			.unwrap()
			.ship;
		let planet_handle = match ship_data.get_data().get_state() {
			ShipState::Landed { target } => *target,
			_ => unreachable!("tried to draw planet interface while not landed!"),
		};
		let planet = input.ct.get_system_object(planet_handle);

		// Reconfigure for new planet if necessary
		if self
			.current_object
			.map(|x| x != planet_handle)
			.unwrap_or(true)
		{
			self.reflow(planet, state);
		}

		// Draw elements
		self.button.push_to_buffer(input, state);
		self.landscape.push_to_buffer(input, state);
		self.frame.push_to_buffer(input, state);
	}

	fn get_textareas(
		&'this self,
		v: &mut Vec<TextArea<'this>>,
		_input: &RenderInput,
		state: &RenderState,
	) {
		v.push(self.description.get_textarea(state));
		v.push(self.title.get_textarea(state));
	}
}
