use galactica_content::{Content, SystemObject, SystemObjectHandle, UiPositionAnchor};
use galactica_system::{data::ShipState, phys::PhysSimShipHandle};
use glyphon::{cosmic_text::Align, Attrs, Color, Metrics, TextArea, Weight};
use nalgebra::{Point2, Vector2};

use crate::{
	ui::{
		manager::{UiScene, UiSceneStepResult, UiScenes},
		util::{SpriteRect, UiSprite, UiTextArea},
	},
	RenderInput, RenderState,
};

use super::UiLandedScene;

pub struct UiOutfitterScene {
	// UI elements
	se_box: UiSprite,
	exit_button: UiSprite,
	exit_text: UiTextArea,

	/// What object we're displaying currently.
	/// Whenever this changes, we need to reflow text.
	current_object: Option<SystemObjectHandle>,

	/// True if we've caught a left click event.
	/// Used for edge detection.
	leftclick_down: bool,
}

impl UiOutfitterScene {
	pub fn new(ct: &Content, state: &mut RenderState) -> Self {
		let exit_button = UiSprite::from(ct, &ct.get_ui().outfitter_exit_button);
		let se_box = UiSprite::from(ct, &ct.get_ui().outfitter_se_box);

		let s = Self {
			// height of element in logical pixels
			current_object: None,
			leftclick_down: false,

			exit_text: UiTextArea::new(
				ct,
				state,
				SpriteRect {
					pos: Point2::new(0.0, 0.0),
					dim: Vector2::new(200.0, 200.0), // TODO: do this better
					anchor_self: UiPositionAnchor::Center,
					anchor_parent: UiPositionAnchor::Center,
				},
				Metrics::new(16.0, 18.0),
				Color::rgb(255, 255, 255),
				Align::Center,
			),

			se_box,
			exit_button,
		};

		return s;
	}

	fn reflow(&mut self, planet: &SystemObject, state: &mut RenderState) {
		self.exit_text.set_text(
			state,
			"Exit",
			Attrs::new()
				.weight(Weight::NORMAL)
				.family(glyphon::Family::SansSerif),
		);

		self.current_object = Some(planet.handle);
	}
}

impl<'this> UiScene<'this> for UiOutfitterScene {
	fn step(&mut self, input: &RenderInput, state: &mut RenderState) -> UiSceneStepResult {
		self.se_box.step(input, state);
		self.exit_button.step(input, state);

		let mut new_scene = None;
		if input.player.input.pressed_leftclick() && !self.leftclick_down {
			self.leftclick_down = true;
			if self.exit_button.contains_mouse(input, state) {
				new_scene = Some(UiScenes::Landed(UiLandedScene::new(input.ct, state)));
			}
		} else if !input.player.input.pressed_leftclick() {
			self.leftclick_down = false;
		}

		return UiSceneStepResult { new_scene };
	}

	fn draw(&mut self, input: &RenderInput, state: &mut RenderState) {
		// Get required data
		let ship_handle = input.player.ship.unwrap();
		let ship_data = &input
			.phys_img
			.get_ship(&PhysSimShipHandle(ship_handle))
			.unwrap()
			.ship;
		let planet_handle = match ship_data.get_data().get_state() {
			ShipState::Landed { target } => *target,
			_ => unreachable!("tried to draw planet interface while not landed!"),
		};
		let planet = input.ct.get_system_object(planet_handle);

		// Reconfigure for new planet if necessary
		if self
			.current_object
			.map(|x| x != planet_handle)
			.unwrap_or(true)
		{
			self.reflow(planet, state);
		}

		// Draw elements
		self.se_box.push_to_buffer(input, state);
		self.exit_button.push_to_buffer(input, state);
	}

	fn get_textareas(
		&'this self,
		v: &mut Vec<TextArea<'this>>,
		_input: &RenderInput,
		state: &RenderState,
	) {
		v.push(self.exit_text.get_textarea(state));
	}
}
