mod sprite;
mod textarea;

use galactica_content::UiPositionAnchor;
pub(super) use sprite::UiSprite;
pub(super) use textarea::UiTextArea;

use nalgebra::{Point2, Vector2};
use winit::dpi::LogicalSize;

use crate::{RenderInput, RenderState};

/// Represents a rectangular region
#[derive(Debug, Clone, Copy)]
pub(crate) struct SpriteRect {
	/// The position of the top-left corner of this rectangle, in fractional units.
	/// (0.0 is left edge of sprite, 1.0 is right edge)
	pub pos: Point2<f32>,

	/// The width and height of this rectangle, in fractional units.
	/// 1.0 will be as tall as the sprite, 0.5 will be half as tall
	pub dim: Vector2<f32>,

	/// How to compute this rectangle's coordinates relative to its parent
	pub anchor_self: UiPositionAnchor,

	/// How to compute this rectangle's coordinates relative to its parent
	pub anchor_parent: UiPositionAnchor,
}

impl SpriteRect {
	/*
		pub fn to_centered_relative(&self, parent: SpriteRectCentered) -> SpriteRectCentered {
			let dim = Vector2::new(self.dim.x * parent.dim.x, self.dim.y * parent.dim.y);

			let pos = Vector2::new(self.pos.x * parent.dim.x, self.pos.y * parent.dim.y);

			let mut zero = match self.anchor_parent {
				PositionAnchor::Center => parent.pos,

				PositionAnchor::NorthWest => Point2::new(
					parent.pos.x - (parent.dim.x / 2.0),
					parent.pos.y + (parent.dim.y / 2.0),
				),

				PositionAnchor::SouthWest => Point2::new(
					parent.pos.x - (parent.dim.x / 2.0),
					parent.pos.y - (parent.dim.y / 2.0),
				),

				PositionAnchor::NorthEast => Point2::new(
					parent.pos.x + (parent.dim.x / 2.0),
					parent.pos.y + (parent.dim.y / 2.0),
				),

				PositionAnchor::SouthEast => Point2::new(
					parent.pos.x + (parent.dim.x / 2.0),
					parent.pos.y - (parent.dim.y / 2.0),
				),
			};

			match self.anchor_self {
				PositionAnchor::Center => {}

				PositionAnchor::NorthWest => {
					zero += Vector2::new(dim.x, -dim.y) / 2.0;
				}

				PositionAnchor::NorthEast => {
					zero += Vector2::new(-dim.x, -dim.y) / 2.0;
				}

				PositionAnchor::SouthWest => {
					zero += Vector2::new(dim.x, dim.y) / 2.0;
				}

				PositionAnchor::SouthEast => {
					zero += Vector2::new(-dim.x, dim.y) / 2.0;
				}
			};

			return SpriteRectCentered {
				dim,
				pos: zero + pos,
			};
		}
	*/

	/// Convert this rectangle to a centered rectangle.
	pub fn to_centered(&self, state: &RenderState) -> CenteredSpriteRect {
		let w: LogicalSize<f32> = state.window_size.to_logical(state.window.scale_factor());
		let w = Vector2::new(w.width, w.height);

		let mut pos = self.pos;
		let dim = self.dim;

		// Origin
		match self.anchor_parent {
			UiPositionAnchor::Center => {}

			UiPositionAnchor::NorthWest => {
				pos += Vector2::new(-w.x, w.y) / 2.0;
			}

			UiPositionAnchor::SouthWest => {
				pos += Vector2::new(-w.x, -w.y) / 2.0;
			}

			UiPositionAnchor::NorthEast => {
				pos += Vector2::new(w.x, w.y) / 2.0;
			}

			UiPositionAnchor::SouthEast => {
				pos += Vector2::new(w.x, -w.y) / 2.0;
			}
		}

		// Offset for self dimensions
		match self.anchor_self {
			UiPositionAnchor::Center => {}

			UiPositionAnchor::NorthWest => {
				pos += Vector2::new(dim.x, -dim.y) / 2.0;
			}

			UiPositionAnchor::NorthEast => {
				pos += Vector2::new(-dim.x, -dim.y) / 2.0;
			}

			UiPositionAnchor::SouthWest => {
				pos += Vector2::new(dim.x, dim.y) / 2.0;
			}

			UiPositionAnchor::SouthEast => {
				pos += Vector2::new(dim.x, dim.y) / 2.0;
			}
		};

		return CenteredSpriteRect { pos, dim };
	}
}

/// Represents a rectangular region, in absolute coordinates relative to the screen center.
#[derive(Debug, Clone, Copy)]
pub(crate) struct CenteredSpriteRect {
	/// The position of the top-left corner of this rectangle, in fractional units.
	/// (0.0 is left edge of sprite, 1.0 is right edge)
	pub pos: Point2<f32>,

	/// The width and height of this rectangle, in fractional units.
	/// 1.0 will be as tall as the sprite, 0.5 will be half as tall
	pub dim: Vector2<f32>,
}

impl CenteredSpriteRect {
	pub fn contains_point(&self, pt: Point2<f32>) -> bool {
		let ne = self.pos + Vector2::new(-self.dim.x, self.dim.y) / 2.0;
		let sw = self.pos + Vector2::new(self.dim.x, -self.dim.y) / 2.0;
		return (pt.y < ne.y && pt.y > sw.y) && (pt.x > ne.x && pt.x < sw.x);
	}
}

pub(super) trait UiElement {
	fn push_to_buffer_child(
		&self,
		input: &RenderInput,
		state: &mut RenderState,
		parent_pos: Point2<f32>,
		parent_size: Vector2<f32>,
	);
}
