use galactica_content::{Content, UiTextAlign, UiTextConfig};
use glyphon::{cosmic_text::Align, Attrs, Buffer, Color, Metrics, Shaping, TextArea, TextBounds};
use nalgebra::Vector2;

use super::SpriteRect;
use crate::RenderState;

/// Represents a text area inside a sprite.
pub(crate) struct UiTextArea {
	/// Bounds of text area
	rect: SpriteRect,

	/// Text buffer
	buffer: Buffer,

	/// Text color
	color: Color,

	/// Text alignment
	align: Align,
}

impl UiTextArea {
	pub fn new(
		_ct: &Content,
		state: &mut RenderState,
		rect: SpriteRect,
		text_metrics: Metrics,
		color: Color,
		align: Align,
	) -> Self {
		let mut s = Self {
			buffer: Buffer::new(&mut state.text_font_system, text_metrics),
			rect,
			align,
			color,
		};

		s.buffer.set_size(
			&mut state.text_font_system,
			s.rect.dim.x * state.window.scale_factor() as f32,
			s.rect.dim.y * state.window.scale_factor() as f32,
		);

		return s;
	}

	pub fn from(ct: &Content, state: &mut RenderState, ui: &UiTextConfig) -> Self {
		Self::new(
			ct,
			state,
			SpriteRect {
				pos: ui.rect.pos,
				dim: ui.rect.dim,
				anchor_self: ui.rect.anchor_self,
				anchor_parent: ui.rect.anchor_parent,
			},
			Metrics::new(ui.font_size, ui.line_height),
			Color::rgb(255, 255, 255),
			match ui.align {
				UiTextAlign::Center => Align::Center,
				UiTextAlign::Left => Align::Left,
			},
		)
	}

	pub fn set_text(&mut self, state: &mut RenderState, text: &str, attrs: Attrs) {
		self.buffer
			.set_text(&mut state.text_font_system, text, attrs, Shaping::Advanced);

		for l in &mut self.buffer.lines {
			l.set_align(Some(self.align));
		}
		self.buffer.shape_until_scroll(&mut state.text_font_system);
	}

	pub fn get_textarea(&self, state: &RenderState) -> TextArea {
		let rect = self.rect.to_centered(state);

		// Glypon works with physical pixels, so we must do some conversion
		let fac = state.window.scale_factor() as f32;
		let corner_ne = Vector2::new(
			(rect.pos.x - rect.dim.x / 2.0) * fac + state.window_size.width as f32 / 2.0,
			state.window_size.height as f32 / 2.0 - (rect.pos.y * fac + rect.dim.y / 2.0),
		);
		let corner_sw = corner_ne + rect.dim * fac;

		TextArea {
			buffer: &self.buffer,
			top: corner_ne.y,
			left: corner_ne.x,
			scale: 1.0,
			bounds: TextBounds {
				top: (corner_ne.y) as i32,
				bottom: (corner_sw.y) as i32,
				left: (corner_ne.x) as i32,
				right: (corner_sw.x) as i32,
			},
			default_color: self.color,
		}
	}
}
